"""
cambridge is a terminal version of Cambridge Dictionary.
The dictionary data comes from https://dictionary.cambridge.org
If you're not satisfied with the result, you can add "-w" flag to look up the word in Merriam-Webster Dictionary.
"""

__version__ = "3.13.11"
